from dnslib import *
from dnslib import server
import time
import sys
#import logging
#import logging.handlers
import copy

local_addr = ''
local_port = 53
resolveLog=False

# Customize the address and port of the external DNS server
class DRecord(DNSRecord):
    @classmethod
    def question(cls,qname,qtype="A",qclass="IN"):
        """
            Shortcut to create question

            >>> q = DNSRecord.question("www.google.com")
            >>> print(q)
            ;; ->>HEADER<<- opcode: QUERY, status: NOERROR, id: ...
            ;; flags: rd; QUERY: 1, ANSWER: 0, AUTHORITY: 0, ADDITIONAL: 0
            ;; QUESTION SECTION:
            ;www.google.com.                IN      A

            >>> q = DNSRecord.question("www.google.com","NS")
            >>> print(q)
            ;; ->>HEADER<<- opcode: QUERY, status: NOERROR, id: ...
            ;; flags: rd; QUERY: 1, ANSWER: 0, AUTHORITY: 0, ADDITIONAL: 0
            ;; QUESTION SECTION:
            ;www.google.com.                IN      NS
        """
        return DRecord(q=DNSQuestion(qname,getattr(QTYPE,qtype),
                                             getattr(CLASS,qclass)))

    def send(self,dest,port=53,tcp=False,timeout=None,ipv6=False):
        """
            Send packet to nameserver and return response
        """
        data = self.pack()
        if ipv6:
            inet = socket.AF_INET6
        else:
            inet = socket.AF_INET
        if tcp:
            if len(data) > 65535:
                raise ValueError("Packet length too long: %d" % len(data))
            data = struct.pack("!H",len(data)) + data
            sock = socket.socket(inet,socket.SOCK_STREAM)
            if timeout is not None:
                sock.settimeout(timeout)
            sock.connect((dest,port))
            sock.sendall(data)
            response = sock.recv(8192)
            length = struct.unpack("!H",bytes(response[:2]))[0]
            while len(response) - 2 < length:
                response += sock.recv(8192)
            sock.close()
            response = response[2:]
        else:
            sock = socket.socket(inet,socket.SOCK_DGRAM)
            if timeout is not None:
                sock.settimeout(timeout)
            sock.sendto(self.pack(),(dest,port))
            response,server = sock.recvfrom(8192)
            sock.close()
        return response

class Resolver:
    def loadConfig(self,dnsConfig,resolveLogger=None):
        self.dnsConfig=dnsConfig
        self.resolveLogger=resolveLogger
        if '*' in dnsConfig:
            self.defserver=dnsConfig['*']['*']
        else:
            self.defserver=['8.8.8.8']
    def spider(self,content):
        if self.resolveLogger:
            self.resolveLogger.debug(content)
    def resolve(self, request, handler):
        d = request.reply()
        q = request.get_q()
        q_name = str(q.qname)
        if q_name=='vpn.by.box.':
            f=os.popen('ip addr show br-lan|grep "inet "')
            LANIP=f.readline()
            LANIP=LANIP.strip().split(' ')[1]
            LANIP=LANIP[:LANIP.find('/')]
            d.add_answer(*RR.fromZone("by.box 1200 A "+LANIP))
            return d
        q_name_begin_with=q_name.split('.')[-3][0]
        Dserver=[]
        if q_name_begin_with in self.dnsConfig:
            for k in self.dnsConfig[q_name_begin_with]:
                if ('.'+q_name)[:-1].endswith('.'+k):
                    Dserver=copy.copy(self.dnsConfig[q_name_begin_with][k])
                    break
        valid=True
        validtime=len(Dserver)
        if validtime==0:
            valid=False
        #if sys.argv[-1]=='local':
            #self.defserver=upstreamServer
        Dserver.extend(upstreamServer)
        Dserver.append('8.8.8.8')
        fail=0
        for i in Dserver:
            try:
                a = DRecord.parse(DRecord.question(q_name).send(i, timeout=3))
            except:
                fail=fail+1
                if fail>=validtime:
                    valid=False
                continue
            for rr in a.rr:
                d.add_answer(rr)
                if isinstance(rr.rdata,dns.A):
                    IP=''
                    for ip in rr.rdata.data:
                        IP=IP+str(ip)+'.'
                    IP=IP[:-1]
                    if valid:
                        os.system('ip route add '+IP+'/32 dev tun0')
                    self.spider(q_name+' '+IP)
                    break
            break
        return d


class pResolver():
    def loadConfig(self,upstreamServer):
        self.upstreamServer=upstreamServer
    def resolve(self, request, handler):
        d = request.reply()
        q = request.get_q()
        q_name = str(q.qname)
        if q_name=='vpn.by.box.':
            f=os.popen('ip addr show br-lan|grep "inet "')
            LANIP=f.readline()
            LANIP=LANIP.strip().split(' ')[1]
            LANIP=LANIP[:LANIP.find('/')]
            d.add_answer(*RR.fromZone(by.box 1200 A "+LANIP))
            return d
        for i in self.upstreamServer:
            try:
                a = DRecord.parse(DRecord.question(q_name).send(i, timeout=2))
            except:
                continue
            for rr in a.rr:
                d.add_answer(rr)
            break
        return d

def getLocalDNS():
    localDNS=[]
    try:
        with open('/tmp/resolv.conf.auto') as temp:
            for line in temp:
                if 'nameserver' in line:
                    line=line.strip()
                    line=line.split(' ')
                    localDNS.append(line[-1])
    except:
        pass
    if localDNS==[]:
        localDNS=['8.8.8.8']
    return localDNS

global upstreamServer
upstreamServer=getLocalDNS()
if len(sys.argv)==1:
    r=pResolver()
    r.loadConfig(upstreamServer)
else:
    config=open('/etc/pyDNSMASQ/config')
    dnsConfig={}
    for line in config:
        line=line.lstrip()
        if line[0]=='#':
            continue
        line=line.strip()
        line=line.split(' ')
        temp=line[0].split('.')
        try:
            if temp[-2][0] not in dnsConfig:
                dnsConfig[temp[-2][0]]={}
            dnsConfig[temp[-2][0]][line[0]]=line[1:]
        except:
            dnsConfig['*']={'*':line[1:]}
    config.close()
    for i in dnsConfig:
        for j in dnsConfig[i]:
            for k in range(0,len(dnsConfig[i][j])):
                dnsConfig[i][j][k]=sys.argv[int(dnsConfig[i][j][k])]
    r = Resolver()
    if resolveLog:
        hdl=logging.handlers.RotatingFileHandler("/etc/pyDNSMASQ/log",maxBytes=100000000,backupCount=4)
        resolveLogger=logging.getLogger('resolveLogger')
        resolveLogger.setLevel('DEBUG')
        formatter=logging.Formatter('%(asctime)s %(filename)s[line:%(lineno)d] %(message)s')
        hdl.setFormatter(formatter)
        resolveLogger.addHandler(hdl)
        r.loadConfig(dnsConfig,resolveLogger)
    else:
        r.loadConfig(dnsConfig)
s = server.DNSServer(r, port=local_port, address=local_addr)
s.start_thread()
try:
    while True:
        time.sleep(10)
        upstreamServer=getLocalDNS()
except:
    print 'Service Stoped'
    sys.exit(0)
